@@ -1,6 +1,7 @@
-"""说明 Introduction:
+"""简介 Introduction:
 A simple text editor written in Python.It supports editing text files,
-binary files ,encodings and changing font size.
+binary files with various encodings which can be automatically detected
+and changing font size.
 When you edit a binary file, the contents of the file are
 displayed as escape sequences.
 You can also find and replace words.
@@ -8,12 +9,13 @@ In addition, code highlighting is supported when editing Python code files,like
 What's more, dragging and dropping files into the editor window is now supported.
 
 一款使用tkinter编写的文本编辑器, 支持编辑文本文件、二进制文件、改变字体大小。
-支持ansi、gbk、utf-8等编码。编辑二进制文件时, 文件内容以转义序列形式显示。
+支持ansi、gbk、utf-8等编码, 以及调用chardet库自动检测编码。
+编辑二进制文件时, 文件内容以转义序列形式显示。
 支持查找、替换、改变字体大小; 且支持撤销、重做; 支持将文件拖放入窗口。
 编辑python代码文件时, 支持代码高亮显示, 类似IDLE。
 
 作者:qfcy (七分诚意)
-版本:1.2.8.8
+版本:%s
 """
 import sys,os,time,pickle
 from tkinter import *
@@ -22,7 +24,9 @@ import tkinter.ttk as ttk
 import tkinter.messagebox as msgbox
 import tkinter.filedialog as dialog
 import tkinter.simpledialog as simpledialog
-
+from tkinter import font
+# 以下为可选(非必需)的模块
+import webbrowser
 try:
     from idlelib.colorizer import ColorDelegator
     from idlelib.percolator import Percolator
@@ -30,10 +34,12 @@ except ImportError:
     ColorDelegator=Percolator=None
 try:import windnd
 except ImportError:windnd=None
+try:import chardet
+except ImportError:chardet=None
 
 __email__="3416445406@qq.com"
 __author__="七分诚意 qq:3076711200 邮箱:%s"%__email__
-__version__="1.2.8.8"
+__version__="1.3.0";__doc__=__doc__%__version__ # 在__doc__中加入版本信息
 
 def view_hex(bytes):
     result=''
@@ -42,12 +48,13 @@ def view_hex(bytes):
         if (i+1) % 4 == 0:result+='\n'
     return result
 
-def to_escape_str(bytes):
+def to_escape_str(bytes,linesep=True):
     # 将字节(bytes)转换为转义字符串
+    # linesep: 是否以length间隔加入换行符, 加入换行符可提高Text控件的显示速度
     str='';length=1024
     for i in range(0,len(bytes),length):
         str+=repr( bytes[i: i+length] ) [2:-1]
-        str+='\n'
+        if linesep:str+='\n'
     return str
 
 def to_bytes(escape_str):
@@ -62,7 +69,7 @@ def to_bytes(escape_str):
         return eval("b'''"+escape_str+"'''")
 
 def bell_(widget=None):
-    try: 
+    try:
         import winsound
         winsound.PlaySound('.',winsound.SND_ASYNC)
     except (ImportError, RuntimeError):
@@ -120,10 +127,15 @@ class SearchDialog(Toplevel):
         text=self.master.contents
         key=self.keyword.get()
         if not key:return
+        # 验证用户输入是否正常
         if self.use_escape_char.get():
             try:key=str(to_bytes(key),encoding=self.coding)
             except Exception as err:
                 handle(err,parent=self);return
+        if self.use_regexpr.get():
+            try:re.compile(key)
+            except re.error:
+                handle(err,parent=self);return
         # 默认从当前光标位置开始查找
         pos=text.search(key,INSERT,END,
                         regexp=self.use_regexpr.get(),
@@ -133,17 +145,22 @@ class SearchDialog(Toplevel):
                         regexp=self.use_regexpr.get(),
                         nocase=not self.match_case.get())
         if pos:
-            newpos="%s+%dc"%(pos,len(key))
+            if self.use_regexpr.get(): # 获取正则表达式匹配的字符串长度
+                text_after = text.get(pos,END)
+                length = re.match(key,text_after).span()[1]
+            else:
+                length = len(key)
+            newpos="%s+%dc"%(pos,length)
             text.mark_set(INSERT,newpos)
             if mark:self.mark_text(pos,newpos)
             return pos,newpos
-        elif bell: # 未找到
+        elif bell: # 未找到,返回None
             bell_(widget=self)
-    def findnext(self,cursor_pos='end',mark=True):
+    def findnext(self,cursor_pos='end',mark=True,bell=True):
         # cursor_pos:标记文本后将光标放在找到文本开头还是末尾
         # 因为search()默认从当前光标位置开始查找
         # end 用于查找下一个操作, start 用于替换操作
-        result=self.search(mark=mark)
+        result=self.search(mark=mark,bell=bell)
         if not result:return
         if cursor_pos=='end':
             self.master.contents.mark_set('insert',result[1])
@@ -152,10 +169,12 @@ class SearchDialog(Toplevel):
         return result
     def mark_text(self,start_pos,end_pos):
         text=self.master.contents
-        text.tag_remove("sel","1.0",END)
-        text.tag_add("sel", start_pos,end_pos)
+        text.tag_remove("sel","1.0",END) # 移除旧的tag
+        # 已知问题: 代码高亮显示时, 无法突出显示找到的文字
+        text.tag_add("sel", start_pos,end_pos) # 添加新的tag "sel"
         lines=text.get('1.0',END)[:-1].count(os.linesep) + 1
         lineno=int(start_pos.split('.')[0])
+        # 滚动文本框, 使被找到的内容显示 ( 由于只判断行数, 已知有bug)
         text.yview('moveto', str((lineno-text['height'])/lines)) # -****- 1.2.5版
         text.focus_force()
         self.master.update_status()
@@ -195,21 +214,22 @@ class ReplaceDialog(SearchDialog):
         options=Frame(self)
         self.create_options(options)
         options.pack(fill=X)
-    def _findnext(self):
+    def _findnext(self):# 仅用于"查找下一个"按钮功能
         text=self.master.contents
         sel_range=text.tag_ranges('sel') # 获得选区的起点和终点
         if sel_range:
             selectarea = sel_range[0].string, sel_range[1].string
             result = self.findnext('start')
+            if result is None:return
             if result[0] == selectarea[0]: # 若仍停留在原位置
-                text.mark_set('insert',result[1])
+                text.mark_set('insert',result[1])# 从选区终点继续查找
                 self.findnext('start')
         else:
             self.findnext('start')
     def replace(self,bell=True,mark=True):
         text=self.master.contents
         result=self.search(mark=False,bell=bell)
-        if not result:return -1 #-1标志已无文本可替换
+        if not result:return # 标志已无文本可替换
         self.master.text_change()
         pos,newpos=result
         newtext=self.text_to_replace.get()
@@ -217,7 +237,9 @@ class ReplaceDialog(SearchDialog):
             newtext=to_bytes(newtext).decode(self.master.coding.get())
         if self.use_regexpr.get():
             old=text.get(pos,newpos)
-            newtext=re.sub(self.keyword.get(),newtext,old)
+            try:newtext=re.sub(self.keyword.get(),newtext,old)
+            except re.error as err:
+                handle(err,parent=self);return
         text.delete(pos,newpos)
         text.insert(pos,newtext)
         end_pos="%s+%dc"%(pos,len(newtext))
@@ -233,26 +255,27 @@ class ReplaceDialog(SearchDialog):
         last = (0,0)
         while True:
             result=self.replace(bell=False,mark=False)
-            if result==-1:break
+            if result is None:break
             flag = True
-            ln,col = self.findnext('start',mark=False)[0].split('.')
-            ln,col = int(ln),int(col)
+            result = self.findnext('start',bell=False,mark=False)
+            if result is None:return
+            ln,col = result[0].split('.')
+            ln = int(ln);col = int(col)
             # 判断新的偏移量是增加还是减小
-            if ln < last[0]:
+            if ln < last[0] or (ln==last[0] and col<last[1]):
                 self.mark_text(*result) # 已完成一轮替换
                 break
-            elif ln==last[0] and col<last[1]:
-                self.mark_text(*result)
-                break
             last=ln,col
         if not flag:bell_()
-     
+
 class Editor(Tk):
     TITLE="PyNotepad"
     encodings="ansi","utf-8","utf-16","utf-32","gbk","big5"
+    # 判断是否有chardet库, 有就启用"自动"功能
+    if chardet is not None:encodings=("自动",)+encodings
     ICON="notepad.ico"
-    NORMAL_CODING="utf-8"
-    FONTSIZES=8, 9, 10, 11, 12, 14, 18, 20, 22, 24, 30
+    NORMAL_CODING="自动" if chardet is not None else "utf-8"
+    FONTSIZES=8, 9, 10, 11, 12, 14, 16, 18, 20, 24, 36, 48
     NORMAL_FONT='宋体'
     NORMAL_FONTSIZE=11
     FILETYPES=[("所有文件","*.*")]
@@ -263,7 +286,7 @@ class Editor(Tk):
     def __init__(self,filename=""):
         super().__init__()
         self.withdraw() # 暂时隐藏窗口,避免调用create_widgets()时窗口闪烁
-        self.title(self.TITLE)
+        self.title(self.TITLE) # 初始化时预先显示标题
         self.bind("<Key>",self.window_onkey)
         self.bind("<FocusIn>",self.focus)
         self.bind("<FocusOut>",self.focus)
@@ -281,6 +304,7 @@ class Editor(Tk):
         self.filename=''
         if filename:
             self.load(filename)
+        else:self.change_title() # 更改标题
     def load_icon(self):
         for path in sys.path + [os.path.split(sys.executable)[0]]: # 用于Py2exe
             try:
@@ -293,8 +317,10 @@ class Editor(Tk):
         self.statusbar.pack(side=BOTTOM,fill=X)
         self.status=Label(self.statusbar,justify=RIGHT)
         self.status.pack(side=RIGHT)
-        self.bin_data=ScrolledText(self.statusbar,width=6,height=6)
-        self.charmap=ScrolledText(self.statusbar,width=14,height=5)
+        self.txt_decoded=ScrolledText(self.statusbar,width=6,height=6)
+        self.txt_decoded.insert('1.0',"在这里查看和编辑解码的数据")
+        self.hexdata=ScrolledText(self.statusbar,width=14,height=5)
+        self.hexdata.insert('1.0',"在这里查看hex十六进制值")
 
         frame=Frame(self)
         frame.pack(side=TOP,fill=X)
@@ -332,16 +358,16 @@ class Editor(Tk):
         self.create_menu()
     def create_binarytools(self):
         if self.isbinary:
-            self.bin_data.pack(side=LEFT,expand=True,fill=BOTH)
-            self.charmap.pack(fill=Y)
+            self.txt_decoded.pack(side=LEFT,expand=True,fill=BOTH)
+            self.hexdata.pack(fill=Y)
             self.status.pack_forget()
             self.status.pack(fill=X)
             self.editmenu.entryconfig(8,state=NORMAL)
         else: # 隐藏工具
-            if self.bin_data:
-                self.bin_data.pack_forget()
-            if self.charmap:
-                self.charmap.pack_forget()
+            if self.txt_decoded:
+                self.txt_decoded.pack_forget()
+            if self.hexdata:
+                self.hexdata.pack_forget()
             self.status.pack(side=RIGHT)
             self.editmenu.entryconfig(8,state=DISABLED)
     def create_menu(self):
@@ -406,22 +432,36 @@ class Editor(Tk):
 
         helpmenu=Menu(self,tearoff=False)
         helpmenu.add_command(label="关于",command=self.about)
+        helpmenu.add_command(label="反馈",command=self.feedback)
 
         menu.add_cascade(label="文件",menu=filemenu)
         menu.add_cascade(label="编辑",menu=self.editmenu)
         menu.add_cascade(label="查看",menu=view)
         menu.add_cascade(label="帮助",menu=helpmenu)
 
-        popup1=Menu(self.bin_data,tearoff=False)
+        # 创建弹出在self.txt_decoded和self.hexdata的菜单
+        popup1=Menu(self.txt_decoded,tearoff=False)
+        def _cut():
+            self.txt_decoded.event_generate("<<Cut>>")
+            self._edit_decoded_event()
+        def _paste():
+            self.txt_decoded.event_generate("<<Paste>>")
+            self._edit_decoded_event()
+        popup1.add_command(label="剪切",command=_cut)
         popup1.add_command(
-            label="复制",command=lambda:self.bin_data.event_generate("<<Copy>>"))
-        popup2=Menu(self.charmap,tearoff=False)
+            label="复制",command=lambda:self.txt_decoded.event_generate("<<Copy>>"))
+        popup1.add_command(label="粘贴",command=_paste)
+
+        popup2=Menu(self.hexdata,tearoff=False)
         popup2.add_command(
-            label="复制",command=lambda:self.charmap.event_generate("<<Copy>>"))
-        self.bin_data.bind("<Button-3>",
+            label="复制",command=lambda:self.hexdata.event_generate("<<Copy>>"))
+
+        self.txt_decoded.bind("<Button-3>",
                     lambda event:popup1.post(event.x_root,event.y_root))
-        self.charmap.bind("<Button-3>",
+        self.txt_decoded.bind("<Key>",self._edit_decoded_event)
+        self.hexdata.bind("<Button-3>",
                     lambda event:popup2.post(event.x_root,event.y_root))
+
         # 显示菜单
         self.config(menu=menu)
 
@@ -443,24 +483,43 @@ class Editor(Tk):
             if dialog.master is self:
                 dialog.findnext()
                 return
+    def _get_fontname(self):
+        font=' '.join(self.contents["font"].split(' ')[:-2])
+        # tkinter会将带空格的字体名称用{}括起来
+        if '{' in font:
+            font = font[1:-1]
+        return font
     def set_fontsize(self,index):
         newsize=self.FONTSIZES[index]
-        fontname=self.contents["font"].split(' ')[0]
+        fontname = self._get_fontname()
         self.contents["font"]=(fontname,newsize,"normal")
     def choose_font(self):
-        oldfont=self.contents["font"].split(' ')[0]
-        font=simpledialog.askstring(
-                '', '输入字体名称: (例如: 楷体)', initialvalue = oldfont)
-        if font is not None:
-            self.contents["font"]=[font] + self.contents["font"].split(' ')[1:]
+        def ok():
+            self.contents["font"]=[opt.get()] + \
+                                   self.contents["font"].split(' ')[-2:] # 保留原先大小、样式
+            dialog.destroy()
+        dialog = Toplevel(self)
+        dialog.title('选择字体')
+        dialog.resizable(False,False)
+        dialog.attributes('-toolwindow',True)
+        opt = ttk.Combobox(dialog)
+        # tkinter.font.families() 获取所有字体名称
+        opt['values']=sorted(font.families())
+        opt.grid(row=0,column=0,columnspan=2,padx=15,pady=20)
+        ttk.Button(dialog,text='确定',command=ok).grid(row=1,column=0)
+        ttk.Button(dialog,text='取消',command=dialog.destroy).grid(row=1,column=1)
+        oldfont = self._get_fontname()
+        opt.set(oldfont)
+        dialog.grab_set() # 对话框打开时, 不允许用户操作主窗口
+        dialog.focus_force()
     def increase_font(self):
         # 增大字体
-        fontsize=int(self.contents["font"].split(' ')[1])
+        fontsize=int(self.contents["font"].split(' ')[-2]) # 使用-2代替1
         index=self.FONTSIZES.index(fontsize)+1
         if 0<=index<len(self.FONTSIZES): self.set_fontsize(index)
     def decrease_font(self):
         # 减小字体
-        fontsize=int(self.contents["font"].split(' ')[1])
+        fontsize=int(self.contents["font"].split(' ')[-2])
         index=self.FONTSIZES.index(fontsize)-1
         if 0<=index<len(self.FONTSIZES): self.set_fontsize(index)
     def set_wrap(self):
@@ -489,6 +548,9 @@ class Editor(Tk):
                 self.decrease_font()
         elif event.keysym.lower()=='f3':
             self.findnext()
+        elif event.keycode == 93: # 按下了菜单键
+            self.editmenu.post(self.winfo_x()+self.winfo_width(),
+                               self.winfo_y()+self.winfo_height())
     def focus(self,event):
         # 当窗口获得或失去焦点时,调用此函数, 用于使对话框置于主窗体前
         for dialog in SearchDialog.instances + ReplaceDialog.instances:
@@ -512,27 +574,34 @@ class Editor(Tk):
         if self.isbinary:
             try:
                 selected=self.contents.get(SEL_FIRST,SEL_LAST)
-                data=to_bytes(selected)
-                try:text=str(data,encoding=self.coding.get(),
+                raw=to_bytes(selected)
+                coding=self.coding.get()
+                # 调用chardet库
+                if coding=="自动":
+                    coding=chardet.detect(raw[:100000])['encoding']
+                    if coding is None:coding='utf-8'
+                try:text=str(raw,encoding=coding,
                              errors="backslashreplace")
                 except TypeError:
                     # 忽略Python 3.4的bug: don't know how to handle
                     # UnicodeDecodeError in error callback
-                    text=str(data,encoding=self.coding.get(),
+                    text=str(raw,encoding=coding,
                              errors="replace")
-                self.bin_data.delete("1.0",END)
-                self.bin_data.insert(INSERT,text)
-                self.charmap.delete("1.0",END)
-                self.charmap.insert(INSERT,view_hex(data))
-                self.status["text"]="选区长度: %d (Bytes)"%len(data)
+                except LookupError as err: # 未知编码
+                    handle(err,parent=self);return
+                self.txt_decoded.delete("1.0",END)
+                self.txt_decoded.insert(INSERT,text)
+                self.hexdata.delete("1.0",END)
+                self.hexdata.insert(INSERT,view_hex(raw))
+                self.status["text"]="选区长度: %d (Bytes)"%len(raw)
             except (TclError,SyntaxError): #忽略未选取内容, 或格式不正确
-                self.bin_data.delete("1.0",END)
-                self.charmap.delete("1.0",END)
+                self.txt_decoded.delete("1.0",END)
+                self.hexdata.delete("1.0",END)
                 self.update_offset()
         else:self.update_offset()
     def update_offset(self,event=None):
         if self.isbinary:
-            prev=self.contents.get("1.0",INSERT)
+            prev=self.contents.get("1.0",INSERT) # 获取从开头到光标处的文本
             try:
                 data=to_bytes(prev)
             except SyntaxError:
@@ -545,8 +614,26 @@ class Editor(Tk):
                 self.status["text"]="偏移量: {} ({})"\
                                      .format(len(data),hex(len(data)))
         else:
-            offset=self.contents.index(CURRENT).split('.')
+            offset=self.contents.index(INSERT).split('.') # 不能用CURRENT
             self.status["text"]="Ln: {}  Col: {}".format(*offset)
+    def _edit_decoded_event(self,event=None):
+        self.after(20,self.edit_decoded) # 如果不使用after(),self.txt_decoded.get不会返回最新的值
+    def edit_decoded(self):
+        range_=self.contents.tag_ranges(SEL) # 获取选区
+        if not range_:return
+        start,end=range_[0].string,range_[1].string # 转换为字符串
+        try:
+            coding=self.coding.get()
+            if coding=="自动":
+                msgbox.showinfo('','不支持自动编码, 请选择或输入其他编码');return
+            byte = self.txt_decoded.get('1.0',END)[:-1].encode(coding)
+            esc_char = to_escape_str(byte,linesep=False)
+            self.file_changed=True;self.change_title()
+            self.contents.delete(start,end)
+            self.contents.insert(start,esc_char)
+            end = '%s+%dc'%(start, len(esc_char))
+            self.contents.tag_add(SEL,start,end)
+        except Exception as err:handle(err)
 
     @classmethod
     def new(cls):
@@ -597,7 +684,7 @@ class Editor(Tk):
             self.file_changed=False
             self.change_title()
             self.change_mode()
-            self.contents.edit_reset() # -*****- 1.2.5版更新 !!!!!!!
+            self.contents.edit_reset() # -*****- 1.2.5版: 重置文本框的撤销功能
             self.contents.focus_force()
         except Exception as err:handle(err,parent=self)
     def _load_data(self,filename):
@@ -609,7 +696,14 @@ class Editor(Tk):
         else:
             try:
                 #读取文件,并对文件内容进行编码
-                data=str(f.read(),encoding=self.coding.get())
+                raw=f.read()
+                if self.coding.get()=="自动":
+                    # 调用chardet库
+                    encoding=chardet.detect(raw[:100000])['encoding']
+                    if encoding is None:
+                        encoding='utf-8'
+                    self.coding.set(encoding)
+                data=str(raw,encoding=self.coding.get())
             except UnicodeDecodeError:
                 f.seek(0)
                 result=msgbox.askyesnocancel("PyNotepad","""%s编码无法解码此文件,
@@ -626,7 +720,7 @@ class Editor(Tk):
             return data
     def insert_hex(self):
         hex = simpledialog.askstring('',
-                    "输入WinHex十六进制数据(如:00 1a 3d ff) :")
+                    "输入WinHex十六进制数据(如:00 1a 3d ff) :",parent=self)
         if hex is None:return
         try:
             data=bytes.fromhex(hex)
@@ -672,7 +766,7 @@ class Editor(Tk):
         if quit:
             Editor.windows.remove(self)
             self.saveconfig()
-            self.destroy()
+            self.destroy() # tkinter不会自动关闭窗口, 需调用函数手动关闭
         return my_ret
     def save(self):
         #保存文件
@@ -683,14 +777,14 @@ class Editor(Tk):
         filename=self.filename
         if filename.strip():
             try:
-                text=self.contents.get('1.0', END)[:-1] # [:-1]: 去除换行符
+                text=self.contents.get('1.0', END)[:-1] # [:-1]: 去除末尾换行符
                 if self.isbinary:
                     data=to_bytes(text)
                 else:
                     data=bytes(text,encoding=self.coding.get(),errors='replace')
-                # Text文本框的bug:避免多余的\r换行符
-                # 如:输入文字foobar, data中变成\rfoobar
-                data=data.replace(b'\r',b'')
+                    # Text文本框的bug:避免多余的\r换行符
+                    # 如:输入文字foobar, data中变成\rfoobar
+                    data=data.replace(b'\r',b'')
                 with open(filename, 'wb') as f:
                     f.write(data)
                 self.filename=filename
@@ -710,6 +804,9 @@ class Editor(Tk):
 
     def about(self):
         msgbox.showinfo("关于",__doc__+"\n作者: "+__author__,parent=self)
+    def feedback(self):
+        webbrowser.open("https://blog.csdn.net/qfcy_/article/details/118399185")
+
     def loadconfig(self):
         try:
             with open(self.CONFIGFILE,'rb') as f:
@@ -718,11 +815,14 @@ class Editor(Tk):
                     setattr(Editor,key,cfg[key])
         except OSError:
             pass
+        # bug修复:未安装chardet时编码设为"自动"的情况
+        if Editor.NORMAL_CODING=="自动" and not chardet:
+            Editor.NORMAL_CODING="utf-8"
     def saveconfig(self):
         font=self.contents['font'].split(' ')
         cfg={'NORMAL_CODING':self.coding.get(),
-             'NORMAL_FONT': font[0],
-             'NORMAL_FONTSIZE': int(font[1]),
+             'NORMAL_FONT': self._get_fontname(),
+             'NORMAL_FONTSIZE': int(font[-2]),
              'AUTOWRAP': self.contents['wrap']}
         with open(self.CONFIGFILE,'wb') as f:
             pickle.dump(cfg,f)
@@ -731,10 +831,10 @@ class Editor(Tk):
         self.after(50,self.onfiledrag2)
     def onfiledrag2(self):
         if not self.contents.get('1.0',END).strip() and not self.file_changed:
-            self.open(self.drag_files[0].decode('gbk'))
+            self.open(self.drag_files[0].decode('ansi'))
             del self.drag_files[0]
         for item in self.drag_files:
-            Editor(item.decode('gbk'))
+            Editor(item.decode('ansi'))
 
 def main():
     if len(sys.argv)>1:
